package de.fraunhofer.iem.app.kpi.service

import de.fraunhofer.iem.app.kpi.dto.KPITreeChildResponseDto
import de.fraunhofer.iem.app.kpi.dto.KPITreeResponseDto
import de.fraunhofer.iem.app.logger.getLogger
import de.fraunhofer.iem.app.repository.dto.RepositoryDetailsDto
import de.fraunhofer.iem.app.tools.occmd.enumeration.Checks
import de.fraunhofer.iem.app.tools.occmd.json.RawResultJson
import de.fraunhofer.iem.kpiCalculator.core.strategy.AggregationKPICalculationStrategy
import de.fraunhofer.iem.kpiCalculator.model.kpi.KpiId
import de.fraunhofer.iem.kpiCalculator.model.kpi.RawValueKpiDto
import org.springframework.stereotype.Service

@Service
class KPIService {

    private val logger = getLogger(javaClass)

    fun removeKPIChildrenLowerThanSecondLevel(kpiTreeResponseDto: KPITreeResponseDto): KPITreeResponseDto {
        val children: List<KPITreeChildResponseDto> = kpiTreeResponseDto.children.map {
            KPITreeChildResponseDto(
                kpi = KPITreeResponseDto(
                    value = it.kpi.value,
                    name = it.kpi.name,
                    description = it.kpi.description,
                    isRoot = it.kpi.isRoot,
                    displayValue = it.kpi.displayValue,
                    emptyList(),
                    it.kpi.isEmpty,
                    order = it.kpi.order
                ),
                actualWeight = it.actualWeight,
                plannedWeight = it.plannedWeight
            )
        }

        return KPITreeResponseDto(
            value = kpiTreeResponseDto.value,
            name = kpiTreeResponseDto.name,
            description = kpiTreeResponseDto.description,
            isRoot = true,
            displayValue = kpiTreeResponseDto.displayValue,
            children = children,
            isEmpty = kpiTreeResponseDto.isEmpty,
            order = kpiTreeResponseDto.order
        )
    }

    /**
     * This method calculates the OccmdCreateKpiDtos based upon the given tool results.
     */
    fun calculateOccmdKpis(rawOccmdResults: List<RawResultJson>): List<RawValueKpiDto> {
        val kpis = mutableListOf<RawValueKpiDto>()

        rawOccmdResults.forEach {
            when (val check = Checks.fromString(it.check)) {
                Checks.CheckedInBinaries ->
                    kpis.add(
                        RawValueKpiDto(
                            kind = KpiId.CHECKED_IN_BINARIES,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.SastUsageBasic ->
                    kpis.add(
                        RawValueKpiDto(
                            kind = KpiId.SAST_USAGE,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.Secrets ->
                    kpis.add(
                        RawValueKpiDto(
                            kind = KpiId.SECRETS,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.CommentsInCode ->
                    kpis.add(
                        RawValueKpiDto(
                            kind = KpiId.COMMENTS_IN_CODE,
                            score = (it.score * 100).toInt()
                        )
                    )

                Checks.DocumentationInfrastructure ->
                    kpis.add(
                        RawValueKpiDto(
                            kind = KpiId.DOCUMENTATION_INFRASTRUCTURE,
                            score = (it.score * 100).toInt()
                        )
                    )

                else ->
                    logger.warn("Unknown OCCMD result. $check")
            }
        }

        return kpis
    }

    /**
     * Creates a named map of RepositoryCreateDtos, based upon the provided repository details.
     * This method only returns raw KPIs.
     */
    fun calculateRepositoryDetailsKpis(repoDetailsDto: RepositoryDetailsDto): List<RawValueKpiDto> {
        return listOf(
            RawValueKpiDto(
                kind = KpiId.NUMBER_OF_COMMITS,
                score = repoDetailsDto.numberOfCommits
            ),
            RawValueKpiDto(
                kind = KpiId.NUMBER_OF_SIGNED_COMMITS,
                score = repoDetailsDto.numberOfSignedCommits
            ),
            RawValueKpiDto(
                kind = KpiId.IS_DEFAULT_BRANCH_PROTECTED,
                score = if (repoDetailsDto.isDefaultBranchProtected) 100 else 0
            )
        )
    }

    fun getKpiTreeForRawKpis(rawKpis: List<KpiCalculationNode>): KpiCalculationNode {
        val rawValueKpis = rawKpis.toMutableList()
        val vulnerabilities = rawValueKpis.filter { it.kind == KpiId.VULNERABILITY_SCORE }
        rawValueKpis.toMutableList().removeAll(vulnerabilities)
        val kpiMap = rawValueKpis.associateBy { it.kind }
        val rootKpi = generateKPITree(kpiMap, vulnerabilities)
        calculateKPIsRecursively(rootKpi)
        return rootKpi
    }

    @Suppress("MagicNumber", "LongMethod")
    private fun generateKPITree(
        rawValueKpis: Map<KpiId, KpiCalculationNode> = emptyMap(),
        vulnerabilityKpis: List<KpiCalculationNode> = emptyList()
    ): KpiCalculationNode {
//        val signedCommitsRatioKPI =
//            KpiCalculationNode(
//                kind = KpiKind.SIGNED_COMMITS_RATIO,
//                calculationStrategy = RatioKPICalculationStrategy()
//            )
//
//        signedCommitsRatioKPI.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.NUMBER_OF_COMMITS,
//                KpiCalculationNode(kind = KpiKind.NUMBER_OF_COMMITS)
//            ),
//            1.0
//        )
//
//        signedCommitsRatioKPI.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.NUMBER_OF_SIGNED_COMMITS,
//                KpiCalculationNode(kind = KpiKind.NUMBER_OF_SIGNED_COMMITS)
//            ),
//            1.0
//        )
//
//        val processComplianceKPI = KpiCalculationNode(
//            kind = KpiKind.PROCESS_COMPLIANCE,
//            calculationStrategy = AggregationKPICalculationStrategy()
//        )
//
//        processComplianceKPI.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.IS_DEFAULT_BRANCH_PROTECTED,
//                KpiCalculationNode(kind = KpiKind.IS_DEFAULT_BRANCH_PROTECTED)
//            ),
//            0.3
//        )
//
//        processComplianceKPI.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.CHECKED_IN_BINARIES,
//                KpiCalculationNode(kind = KpiKind.CHECKED_IN_BINARIES)
//            ),
//            0.2
//        )
//
//        processComplianceKPI.addChildKPI(signedCommitsRatioKPI, 0.2)
//
//        val documentationKpi = KpiCalculationNode(
//            kind = KpiKind.DOCUMENTATION,
//            calculationStrategy = AggregationKPICalculationStrategy()
//        )
//
//        documentationKpi.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.DOCUMENTATION_INFRASTRUCTURE,
//                KpiCalculationNode(kind = KpiKind.DOCUMENTATION_INFRASTRUCTURE)
//            ),
//            0.6
//        )
//
//        documentationKpi.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.COMMENTS_IN_CODE,
//                KpiCalculationNode(kind = KpiKind.COMMENTS_IN_CODE)
//            ),
//            0.4
//        )
//
//        val processTransparencyKPI = KpiCalculationNode(
//            kind = KpiKind.PROCESS_TRANSPARENCY,
//            calculationStrategy = AggregationKPICalculationStrategy()
//        )
//        processComplianceKPI.addChildKPI(documentationKpi, 0.3)
//        processTransparencyKPI.addChildKPI(signedCommitsRatioKPI, 1.0)
//
//        val securityKPI = KpiCalculationNode(
//            kind = KpiKind.SECURITY,
//            calculationStrategy = AggregationKPICalculationStrategy()
//        )
//
//        val internalQuality = KpiCalculationNode(
//            kind = KpiKind.INTERNAL_QUALITY,
//            calculationStrategy = AggregationKPICalculationStrategy()
//        )
//
//        val externalQuality = KpiCalculationNode(
//            kind = KpiKind.EXTERNAL_QUALITY,
//            calculationStrategy = AggregationKPICalculationStrategy()
//        )
//
//        externalQuality.addChildKPI(documentationKpi, 1.0)
//        internalQuality.addChildKPI(documentationKpi, 1.0)
//
//        val maximalDependencyVulnerabilityKPI = KpiCalculationNode(
//            kind = KpiKind.MAXIMAL_VULNERABILITY,
//            calculationStrategy = MaximumKPICalculationStrategy()
//        )
//
//        vulnerabilityKpis.forEach { maximalDependencyVulnerabilityKPI.addChildKPI(it, 0.0) }
//
//        securityKPI.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.SECRETS,
//                KpiCalculationNode(kind = KpiKind.SECRETS)
//            ),
//            0.3
//        )
//
//        securityKPI.addChildKPI(
//            rawValueKpis.getOrDefault(
//                KpiKind.CHECKED_IN_BINARIES,
//                KpiCalculationNode(kind = KpiKind.CHECKED_IN_BINARIES)
//            ),
//            0.2
//        )
//
//        securityKPI.addChildKPI(maximalDependencyVulnerabilityKPI, 0.5)
//
        val rootKPI = KpiCalculationNode(
            kind = KpiId.ROOT,
            calculationStrategy = AggregationKPICalculationStrategy(),
        )

//        rootKPI.addChildKPI(processTransparencyKPI, 0.1)
//        rootKPI.addChildKPI(processComplianceKPI, 0.1)
//        rootKPI.addChildKPI(internalQuality, 0.15)
//        rootKPI.addChildKPI(externalQuality, 0.25)
//        rootKPI.addChildKPI(securityKPI, 0.4)
        return rootKPI
    }

    private fun calculateKPIsRecursively(
        kpi: KpiCalculationNode,
        visited: MutableSet<KpiCalculationNode> = mutableSetOf()
    ) {
        // Check if the KPI has already been visited
        if (visited.contains(kpi)) {
            return
        }

        // Check if the KPI has child KPIs
        if (!kpi.hasChildren()) {
            // Leaf node, calculate the KPI value
            kpi.applyStrategy()
            visited.add(kpi)
            return
        }

        // Recursively calculate child KPIs first
        for (childEdge in kpi.hierarchyEdges) {
            calculateKPIsRecursively(childEdge.to, visited)
        }

        // Calculate the KPI value after processing child KPIs
        kpi.applyStrategy()
        visited.add(kpi)
    }
}
