package de.fraunhofer.iem.kpiCalculator.core

import de.fraunhofer.iem.kpiCalculator.core.hierarchy.KpiCalculationNode
import de.fraunhofer.iem.kpiCalculator.model.kpi.KpiId
import de.fraunhofer.iem.kpiCalculator.model.kpi.KpiStrategyId
import de.fraunhofer.iem.kpiCalculator.model.kpi.RawValueKpiDto
import de.fraunhofer.iem.kpiCalculator.model.kpi.hierarchy.KpiHierarchy
import de.fraunhofer.iem.kpiCalculator.model.kpi.hierarchy.KpiNode
import de.fraunhofer.iem.kpiCalculator.model.kpi.hierarchy.KpiResultNode

sealed class KpiResult {}

object KpiCalculator {
    //XXX: Setup Logger

    fun calculateKpis(hierarchy: KpiHierarchy, rawValueKpiDtos: List<RawValueKpiDto>): KpiResultNode {
        val root = hierarchy.rootNode
        val connectedHierarchyRoot = connectKpiHierarchyToRawValues(root, rawValueKpiDtos)
        depthFirstTraversal(connectedHierarchyRoot) { it.calculateKpi() }
        return KpiCalculationNode.to(connectedHierarchyRoot)
    }

    private fun depthFirstTraversal(
        node: KpiCalculationNode,
        seen: MutableSet<KpiCalculationNode> = mutableSetOf(),
        action: (node: KpiCalculationNode) -> Unit
    ) {
        if (seen.contains(node)) {
            return
        }

        node.hierarchyEdges
            .forEach { child ->
                depthFirstTraversal(node = child.target, seen = seen, action)
            }

        action(node)
        seen.add(node)
    }

    private fun connectKpiHierarchyToRawValues(
        node: KpiNode,
        rawValueKpiDtos: List<RawValueKpiDto>
    ): KpiCalculationNode {

        val kindToValues = mutableMapOf<KpiId, MutableList<RawValueKpiDto>>()
        KpiId.entries.forEach { kindToValues[it] = mutableListOf() }

        rawValueKpiDtos.forEach {
            kindToValues[it.kind]!!.add(it)
        }

        val calculationRoot = KpiCalculationNode.from(node)

        depthFirstTraversal(
            node = calculationRoot
        ) { currentNode ->
            // TODO: this is complicated as we don't know how many children we will have.
            // this scenario might reappear again so we should come up with a better solution
            if (
                currentNode.kind == KpiId.MAXIMAL_VULNERABILITY &&
                currentNode.hierarchyEdges.size == 1 &&
                currentNode.hierarchyEdges.first().target.kind == KpiId.VULNERABILITY_SCORE &&
                !kindToValues[KpiId.VULNERABILITY_SCORE].isNullOrEmpty()
            ) {
                val vulnerabilityCalcNodes = kindToValues[KpiId.VULNERABILITY_SCORE]?.map {
                    val newNode = KpiCalculationNode(
                        kind = KpiId.VULNERABILITY_SCORE,
                        calculationStrategy = KpiStrategyId.RAW_VALUE_STRATEGY
                    )
                    newNode.setScore(it.score)
                    newNode
                }

                if (!vulnerabilityCalcNodes.isNullOrEmpty()) {
                    currentNode.clearChildren()
                    val weights = 1.0 / vulnerabilityCalcNodes.size
                    vulnerabilityCalcNodes.forEach {
                        currentNode.addChild(it, weights)
                    }

                }
            } else {
                val correspondingRawValue = kindToValues[currentNode.kind]
                if (!correspondingRawValue.isNullOrEmpty()) {
                    if (correspondingRawValue.size > 1) {
                        println("We only expect one RAW value for every KPI")
                    }
                    currentNode.setScore(correspondingRawValue.first().score)
                }
            }

        }
        return calculationRoot
    }
}
