package de.fraunhofer.iem.kpiCalculator.core.strategy

import de.fraunhofer.iem.kpiCalculator.model.kpi.hierarchy.KpiCalculationResult


object AggregationKPICalculationStrategy {
    /**
     * This function calculates the aggregate sum of all given children.
     * If a child is empty it is removed from the calculation and its
     * corresponding edge weight is distributed evenly between the
     * remaining children.
     * The method returns the KPIs value as well as the updated
     * KPIHierarchyEdgeDtos with the actual used weight.
     */
    fun calculateKPI(childScores: List<Pair<KpiCalculationResult, Double>>): KpiCalculationResult {

        if (childScores.isEmpty()) {
            return KpiCalculationResult.Empty
        }

        val successScores = childScores.filter { it.first is KpiCalculationResult.Success }
        val failed = childScores
            .filter { it.first !is KpiCalculationResult.Success }
        val missingEdgeWeights = failed.sumOf { it.second }

        val additionalWeight =
            if (missingEdgeWeights == 0.0)
                0.0
            else
                missingEdgeWeights / successScores.size

        val aggregation =
            if (successScores.isEmpty())
                0
            else
                successScores.sumOf {
                    (it.first as KpiCalculationResult.Success).score * (it.second + additionalWeight)
                }.toInt()


        if (failed.isNotEmpty()) {
            return KpiCalculationResult.Incomplete(
                score = aggregation,
                reason = "There were ${failed.size} elements missing during aggregation.",
                missingWeights = missingEdgeWeights
            )
        }

        return KpiCalculationResult.Success(score = aggregation)
    }
}
